/*
 *     Repository.java  Dec 15, 2004
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 *
 *     Thanks to Daichi Oohashi who provides a patch so as to
 *     write well-formed XML.
 */

package org.jtool.platform.j2xml;
import org.jtool.platform.xml.wrapper.*;
import org.jtool.platform.xml.dom.*;
import org.jtool.platform.util.*;
import org.jtool.platform.event.*;
import java.util.*;
import java.io.*;
import org.w3c.dom.*;

public class Repository {
    private static ArrayList javaTexts = new ArrayList();  // JavaXMLText
    private static ArrayList classTexts = new ArrayList();  // ClassXMLText
    private static String dtdName = null;

    public Repository() {
    }

    public static void init() {
        javaTexts.clear();
        classTexts.clear();

        Parser.init();

        ArrayList dirtyFiles = cleanUpSourceXMLDir();
        recoverJavaXMLTexts();

        dirtyFiles = removeDirtyFiles(dirtyFiles);
        setDirtyFilesInRepository(dirtyFiles);
    }

    public static void reinit() {
        ArrayList dirtyFiles = cleanUp();
        setDirtyFilesInRepository(dirtyFiles);
    }

    public static ArrayList cleanUpSourceXMLDir() {
        String sourceDir = SystemProperties.getSourceDir();
        String sourceXMLDir = SystemProperties.getSourceXMLDir();
        ArrayList files = SystemProperties.collectProjectXMLFiles(sourceXMLDir);

        ArrayList dirtyFiles = new ArrayList();
        
        Iterator it = files.iterator();
        while (it.hasNext()) {
            String xmlName = (String)it.next();

            String javaName = sourceDir + File.separator
                + SystemProperties.getRelativePath(xmlName, sourceXMLDir);
            javaName = javaName.substring(0, javaName.lastIndexOf("."));

            File xmlFile = new File(xmlName);
            File javaFile = new File(javaName);

            if (!javaFile.exists() || javaFile.lastModified() > xmlFile.lastModified()) {
                dirtyFiles.add(javaName);
                xmlFile.delete();

                EventManager.getInstance().fire(
                  new NoticeEvent(new Repository(), "DELETE: " + xmlFile.getPath()));
            }
        }
        return dirtyFiles;
    }

    private static ArrayList cleanUp() {
        ArrayList dirtyFiles = new ArrayList();

        Iterator it = javaTexts.iterator();
        while (it.hasNext()) {
            JavaXMLText text = (JavaXMLText)it.next();

            if (text.isDirty()) {
                dirtyFiles.add(text.getName());
            }
        }

        dirtyFiles = getDirtyFiles(dirtyFiles);
        return dirtyFiles;
    }

    public static void recoverJavaXMLTexts() {
        String sourceDir = SystemProperties.getSourceDir();
        ArrayList files = SystemProperties.collectProjectJavaFiles(sourceDir);

        Iterator it = files.iterator();
        while (it.hasNext()) {
            String pathName = (String)it.next();
            recoverJavaXMLText(pathName);
        }

        it = javaTexts.iterator();
        while (it.hasNext()) {
            JavaXMLText text = (JavaXMLText)it.next();
            text.convertToAbsolutePath();
            text.recoverResolvedFqn();
            text.recoverAncestors();
            // text.recoverDescendants();
        }
    }

    private static void recoverJavaXMLText(String pathName) {
        try {
            File xmlFile = new File(SystemProperties.getXMLFileNameInSourceXMLDir(pathName));
            if (xmlFile.canRead()) {
                String content = ReadFile.read(xmlFile);

                JavaXMLText text = new JavaXMLText(pathName, content);
                if (text != null) {
                    add(text);

                    EventManager.getInstance().fire(
                      new NoticeEvent(new Repository(), "ENTRY: " + pathName));
                }
            }
        } catch (Exception e) { }
    }

    public static ArrayList removeDirtyFiles(ArrayList dfiles) {
        ArrayList pfiles = getDirtyFiles(dfiles);

        Iterator it = pfiles.iterator();
        while (it.hasNext()) {
            String pathName = (String)it.next();

            if (!dfiles.contains(pathName)) {
                File xmlFile = new File(SystemProperties.getXMLFileNameInSourceXMLDir(pathName));
                xmlFile.delete();

                EventManager.getInstance().fire(
                  new NoticeEvent(new Repository(), "DELETE: " + xmlFile.getPath()));
            }
        }
        return pfiles;
    }

    public static void setDirtyFilesInRepository(ArrayList dfiles) {
        Iterator it = javaTexts.iterator();
        while (it.hasNext()) {
            JavaXMLText text = (JavaXMLText)it.next();

            if (dfiles.contains(text.getName())) {
                text.setDirty();
            }                
        }
    }

    private static ArrayList getDirtyFiles(ArrayList dfiles) {
        ArrayList pfiles = new ArrayList(dfiles);

        while (true) {
            ArrayList files = getRelatedDirtyFiles(pfiles);
            if (files.size() == 0) {
                break;
            }
            pfiles.addAll(files);
        }
        return pfiles;
    }

    private static ArrayList getRelatedDirtyFiles(ArrayList pfiles) {
        ArrayList files = new ArrayList();

        Iterator it = javaTexts.iterator();
        while (it.hasNext()) {
            JavaXMLText text = (JavaXMLText)it.next();
            JavaFile jf = new JavaFile(text.getDocument().getDocumentElement());

            if (contains(jf.getUsedFiles(), pfiles) && !pfiles.contains(jf.getName())) {
                files.add(jf.getName());
            }
        }
        return files;
    }

    private static boolean contains(ArrayList names1, ArrayList names2) {
        Iterator it = names2.iterator();
        while (it.hasNext()) {
            String name = (String)it.next();

            if (names1.contains(name)) {
                return true;
            }
        }
        return false;
    }

    public static void terminate() {
        SystemProperties.removeDir(SystemProperties.getSourceXMLDir());
        writeJavaXMLText(javaTexts);

        terminateWithoutWriting();
    }

    public static void terminateWithoutWriting() {
        SystemProperties.storeProjectProperties();
        Parser.terminate();
    }

    public static ArrayList getJavaTexts() {
        return javaTexts;
    }

    public static ArrayList getClassTexts() {
        return classTexts;
    }

    public static XMLText getXMLText(String pathName) {
        if (pathName == null){
            return null;
        }

        if (pathName.endsWith(".java")) {
            return getJavaXMLText(pathName);
        } else {
            return getClassXMLText(pathName);
        }
    }
            
    public static JavaXMLText getJavaXMLText(String pathName) {
        if (!pathName.startsWith("/")) {
            pathName = SystemProperties.getSourceDir() + File.separator + pathName;
        }
        return (JavaXMLText)checkXMLText(javaTexts, pathName);
    }

    public static ClassXMLText getClassXMLText(String pathName) {
        return (ClassXMLText)checkXMLText(classTexts, pathName);
    }

    private static XMLText checkXMLText(ArrayList texts, String pathName) {
        Iterator it = texts.iterator();
        while (it.hasNext()) {
            XMLText text = (XMLText)it.next();
            if (text.isTheSameName(pathName)) {
                return text;
            }
        }
        return null;
    }

    public static String getJavaXMLContent(String pathName) {
        JavaXMLText text = entryJavaXMLText(pathName);
        return text.getContent();
    }

    public static Document getJavaXMLDocument(String pathName) {
        JavaXMLText text = entryJavaXMLText(pathName);
        return text.getDocument();
    }

    public static JavaFile getJavaFile(String pathName) {
        XMLText text = getXMLText(pathName);
        if (text != null){
            return new JavaFile(text.getDocument().getDocumentElement());
        }
        return null;
    }

    public static JavaClass getJavaClass(String fqn, JavaFile jfile) {
        JavaClass jc = getJavaClass(fqn);
        if (jc == null) {
            XMLText text = Repository.getXMLText(jfile.getName());

            if (text != null) {
                String pathName = text.getClassFileName(fqn);
                jc = Repository.getJavaClass(fqn, pathName);
            }
        }
        return jc;
    }

    protected static JavaClass getJavaClass(String fqn) {
        String fqnPrefix = fqn;
        int sep = fqn.indexOf("[");
        if (sep != -1) {
            fqnPrefix = fqnPrefix.substring(0, sep);
        }

        String pathName = ClassPathFile.getClassFileName(fqnPrefix);
        return getJavaClass(fqn, pathName);
    }

    protected static JavaClass getJavaClass(String className, String pathName) {
        XMLText text = getXMLText(pathName);
        if (text != null) {
            JavaClass jc = text.getJavaClass(className);
            if (jc != null) {
                return jc;
            }
        }
        return null;
    }

    public static JavaClass entryClass(String className) {
        String pathName = ClassPathFile.getClassFileName(className);
        return entryClass(className, pathName);
    }

    public static JavaClass entryClass(String className, String pathName) {
        JavaClass jc = getJavaClass(className, pathName);
        if (jc != null) {
            return jc;
        }

        ClassXMLText text = entryClassXMLText(pathName);
        if (text != null){
            return text.getJavaClass(className);
        }
        return null;
    }

    public static void add(JavaXMLText text) {
        javaTexts.add(text);
    }

    public static void add(ClassXMLText text) {
        classTexts.add(text);
    }

    public static void remove(ArrayList texts) {
        ArrayList tmpTexts = new ArrayList(texts);

        Iterator it = tmpTexts.iterator();
        while (it.hasNext()) {
            XMLText text = (XMLText)it.next();
            remove(text);
        }
    }

    public static void remove(XMLText text) {
        remove(text.getName());
    }

    public static void remove(JavaXMLText text) {
        javaTexts.remove(text);
    }

    public static void remove(ClassXMLText text) {
        classTexts.remove(text);
    }

    public static void remove(String pathName) {
        JavaXMLText jtext = getJavaXMLText(pathName);
        if (jtext != null) {
            remove(jtext);
        } else {
            ClassXMLText ctext = getClassXMLText(pathName);
            if (ctext != null) {
                remove(ctext);
            }
        }
    }

    public static void clearJavaXMLTexts() {
        clear(javaTexts);
    }

    public static void clearClassXMLTexts() {
        clear(classTexts);
    }

    private static void clear(ArrayList texts) {
        Iterator it = texts.iterator();
        while (it.hasNext()) {
            XMLText text = (XMLText)it.next();
            text = null;
        }
        texts.clear();
    }

    public static ArrayList parse() {
        // System.out.println("Initialize: " + System.currentTimeMillis());
        String sourceDir = SystemProperties.getSourceDir();
        ArrayList files = SystemProperties.collectProjectJavaFiles(sourceDir);

        // System.out.println("Parse (start): " + System.currentTimeMillis());
        ArrayList texts = parse(files);
        // System.out.println("Terninamtes: " + System.currentTimeMillis());
        return texts;
    }
    
    public static ArrayList parse(ArrayList files) {
        Iterator it = files.iterator();
        while (it.hasNext()) {
            String pathName = (String)it.next();
            entryJavaXMLText(pathName);
        }

        cleanUpRepository();
        checkJavaClasses();

        ArrayList texts = Resolver.resolveFqn();
        return texts;
    }

    private static void cleanUpRepository() {
        ArrayList files = new ArrayList(javaTexts);

        Iterator it = files.iterator();
        while (it.hasNext()) {
            JavaXMLText text = (JavaXMLText)it.next();
            if (text.isDirty()) {
                remove(text);
            }
        }
    }

    private static JavaXMLText entryJavaXMLText(String pathName) {
        JavaXMLText text = getJavaXMLText(pathName);

        if (text == null || text.isDirty()) {
            try {
                if (text != null) {
                    remove(text);
                }

                String content;
                if (text != null && text.isOnEdit()){
                    String code = text.getSourceCodeOnEdit();
                    content = Parser.parse(pathName, code);
                } else {
                    content = Parser.parse(pathName);
                }

                if (content != null) {
                    text = new JavaXMLText(pathName, content);

                    if (text != null) {
                        Parser.fix(text.getDocument());
                        Resolver.resolveFileInfo(text);
                        add(text);
                    }
                }
            } catch (Exception e) { }
        }
        return text;
    }

    private static ClassXMLText entryClassXMLText(String pathName) {
        ClassXMLText text = getClassXMLText(pathName);

        if (text == null) {
            try {
                File xmlFile = new File(SystemProperties.getXMLFileNameInClassXMLDir(pathName));
                if (xmlFile.canRead()) {
                    String content = ReadFile.read(xmlFile);
                    text = new ClassXMLText(pathName, content);

                    if (text != null) {
                        add(text);
                        text.convertToAbsolutePath();
                        text.recoverResolvedFqn();
                        text.recoverAncestors();
                        // text.recoverDescendants();

                        // EventManager.getInstance().fire(
                        //   new NoticeEvent(new Repository(), "ENTRY: " + pathName));
                    }

                } else {
                    String content = ClassFileParser.createXMLText(pathName);
                    if (content != null) {
                        text = new ClassXMLText(pathName, content);

                        if (text != null) {
                            AncestorResolver.setAncestors(text);
                            TypeResolver.writeResolvedFqn(text);
                            add(text);
                            writeXMLText(text, xmlFile.getPath());

                            // EventManager.getInstance().fire(
                            //   new NoticeEvent(new Repository(), "WRITE: " + xmlFile.getPath()));
                        }
                    }
                }
            } catch (Exception e) { }
        }
        return text;
    }

    public static void checkJavaClasses() {
        JavaClassList classes = new JavaClassList();

        Iterator it = getJavaTexts().iterator();
        while (it.hasNext()) {
            JavaXMLText text = (JavaXMLText)it.next();

            JavaFile jf = text.getJavaFile();
            checkJavaClasses(jf, classes);
        }

        it = Repository.getClassTexts().iterator();
        while (it.hasNext()) {
            ClassXMLText text = (ClassXMLText)it.next();

            JavaFile jf = text.getJavaFile();
            checkJavaClasses(jf, classes);
        }
    }

    private static void checkJavaClasses(JavaFile jfile, JavaClassList classes) {
        Iterator it = jfile.getAllClassesAndInterfaces().iterator();
        while (it.hasNext()) {
            JavaClass jc = (JavaClass)it.next();
            if (classes.contains(jc)) {
                EventManager.getInstance().fire(new WarningEvent(new Repository(),
                  "The same class " + jc.getFqn() + " already exists in the project"));
            }
            classes.add(jc);
        }
    }

    public static void writeJavaXMLText(ArrayList texts) {
        Iterator it = texts.iterator();
        while (it.hasNext()) {
            JavaXMLText text = (JavaXMLText)it.next();
            String xmlName = SystemProperties.getXMLFileNameInSourceXMLDir(text.getName());
            writeXMLText(text, xmlName);
        }
    }

    public static void writeXMLText(XMLText text, String xmlName) {
        text.convertToRelativePath();
        String content = text.getCurrentContent(dtdName);
        String dirName = xmlName.substring(0, xmlName.lastIndexOf(File.separator));
        SystemProperties.makeDir(dirName);

        try {
            WriteFile.write(xmlName, content);
        } catch (IOException e) { }
    }

    public static void setLineNumber() {
        Iterator it = javaTexts.iterator();
        while (it.hasNext()) {
            JavaXMLText text = (JavaXMLText)it.next();
            Resolver.resolveLineNumber(text);
        }
    }

    public static void setDescendants() {
        Iterator it = javaTexts.iterator();
        while (it.hasNext()) {
            JavaXMLText text = (JavaXMLText)it.next();
            AncestorResolver.setDescendants(text);
        }
    }

    public static void setCompilerCheck(boolean bool) {
        Parser.setCompilerCheck(bool);
    }

    public static boolean addEventListener(WarningEventListener l) {
        return EventManager.getInstance().addEventListener(l);
    }

    public static boolean addEventListener(LogEventListener l) {
        return EventManager.getInstance().addEventListener(l);
    }

    public static boolean addEventListener(NoticeEventListener l) {
        return EventManager.getInstance().addEventListener(l);
    }

    public static boolean removeEventListener(WarningEventListener l) {
        return EventManager.getInstance().removeEventListener(l);
    }

    public static boolean removeEventListener(LogEventListener l) {
        return EventManager.getInstance().removeEventListener(l);
    }

    public static boolean removeEventListener(NoticeEventListener l) {
        return EventManager.getInstance().removeEventListener(l);
    }

    public static void printAllXMLDocuments() {
        Iterator it = javaTexts.iterator();
        while (it.hasNext()) {
            JavaXMLText text = (JavaXMLText)it.next();
            String content = text.getCurrentContent(dtdName);

            System.out.println("----");
            System.out.print(content);
            System.out.println("----");
        }
    }

    public static void visitAllXMLDocuments() {
        DOM2Code.initNum();
        Iterator it = javaTexts.iterator();
        while (it.hasNext()) {
            JavaXMLText text = (JavaXMLText)it.next();
            DOM2Code.visitNode(text.getDocument());
        }
        System.out.println(".java");
        DOM2Code.termNum();

        /*
        DOM2Code.initNum();
        it = classTexts.iterator();
        while (it.hasNext()) {
            ClassXMLText text = (ClassXMLText)it.next();
            DOM2Code.visitNode(text.getDocument());
        }
        System.out.println(".class");
        DOM2Code.termNum();
        */
    }

    public static void main(String[] args) {
        if (args.length >= 3) {
            Options.parse(args);
            String sapidDir = Options.getOption("-sapiddir");
            String cygwinDir = Options.getOption("-cygwindir");
            String classPath = Options.getOption("-classpath");
            dtdName = Options.getOption("-dtd");

            if (sapidDir != null) {
                String arg;
                for (int i = 0; (arg = Options.getFile(i)) != null; i++) {

                    String fileName = SystemProperties.getCanonicalPath(arg);
                    String dirName;

                    File file = new File(fileName);
                    if (file.isDirectory()) {
                        dirName = fileName;
                        fileName = null;
                    } else {
                        dirName = file.getParent();
                    }

                    SystemProperties.setApplicationDir(".");
                    SystemProperties.setSapidDir(sapidDir);
                    SystemProperties.setCygwinDir(cygwinDir);
                    SystemProperties.setDefaultProjectSetting(dirName);

                    if (classPath != null) {
                        SystemProperties.setClassPath(dirName + File.pathSeparator + classPath);
                    } else {
                        SystemProperties.setClassPath(dirName);
                    }

                    long startTime, endTime;

                    startTime = System.currentTimeMillis();
                    Repository.init();

                    if (fileName == null) {
                        Repository.parse();
                    } else {
                        ArrayList files = new ArrayList();
                        files.add(fileName);
                        Repository.parse(files);
                    }

                    endTime = System.currentTimeMillis();

                    long convertTime = endTime - startTime;
                    System.out.print("Conversion Time: " + convertTime + " [ms]");
                    System.out.println("(start: " + startTime + " end: " + endTime + ")");

                    // Repository.printAllXMLDocuments();

                    /*
                    startTime = System.currentTimeMillis();
                    Repository.visitAllXMLDocuments();
                    endTime = System.currentTimeMillis();

                    long visitTime = endTime - startTime;
                    System.out.print("Visit Time: " + visitTime + " [ms]");
                    System.out.println("(start: " + startTime + " end: " + endTime + ")");
                    */
                }

                Repository.terminate();
                System.exit(0);
            }
        }

        System.err.println("Usage: java Repository -sapiddir sapid_dir [ -classpath additional_classpath ] project_top_dir");
    }
}
